//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
//  main.cpp
//  This is a part of the lslua source code.
//
//  Copyright (C) 2003 ilmcuts.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

#include "common.hpp"
#include "exports.hpp"
#include "scripting.h"
#include "enforce_simple.hpp"

#include "main.h"

using namespace lssdk;

const char rcsRevision[] = "0.3.1";
const char aboutString[] = "LSLua 0.3.1";


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// typedefs
//
typedef bool (*ConfigCallback)(LPCSTR pszFile);

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Globals
//
Lua* lua_state;
std::vector< std::string > loaded_bangs;

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Error
//
void Error(const char* pszFormat, ...)
{
    char szMessage[MAX_ERROR] = { 0 };
    va_list argList;
    
    va_start(argList, pszFormat);
    StringCchVPrintf(szMessage, MAX_ERROR, pszFormat, argList);
    va_end(argList);
    
    MessageBox(NULL, szMessage, "LSLua Error", MB_TOPMOST | MB_SETFOREGROUND | MB_ICONERROR);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// HandleLua
//
//bool HandleLua(LPCSTR pszConfig)
//{
    //
    // *Lua function custombang()
    // *Lua   print("Wheeee")
    // *Lua end
    // *Lua AddBang("luatest", custombang)
    //
//    return false;
//}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// createVarTable
//
/*
void createVarTable()
{
    // load the lua script which creates our evar table
    char buf[ MAX_LINE_LENGTH ] = { 0 };
    VarExpansion( buf, "$ModulesDir$" );
    std::string path = buf;
    path += "lslua\\vars.lua";
    lua_state->add_script( path.c_str() );
    // and call the function that sets up the var table
    lua_state->trigger( "_ls_init" );
/ *
    lua_State* state = lua_state->get_lua_state();

    // create the table
    // stack: [ ls<table> ]
    lua_newtable( state );

    // override the metatable functions
    // --------------------
    
    // get the metatable for this table
    // stack: [ ls<table>, ls.mt<table> ]
    lua_getmetatable( state, 1 );

    // push the metatable key onto the stack first
    // stack: [ ls<table>, ls.mt<table>, "__index"<str> ]
    lua_pushstring( state, "__index" );

    // then get the lua function for indexing on top of the stack
    // stack: [ ls<table>, ls.mt<table>, "__index"<str>, _ls_index<func> ]
    lua_pushstring( state, "_ls_index" );
    lua_rawget( state, LUA_GLOBALSINDEX );
    if ( !lua_isfunction( state, 1 ) ) {
        lua_state->error( "_ls_index is not a function" );
        return;
    }

    // set _ls_index as __index for this table
    // stack: [ ls<table>, ls.mt<table> ]
    lua_rawset( state, 3 );

    // get rid of the unneeded metatable
    // stack: [ ls<table> ]
    lua_pop( state, 1 );

    // register the table in the global environment
    // ----------------------

    // push the key
    // stack: [ ls<table>, "ls"<str> ]
    lua_pushstring( state, "ls" );

    // swap the values, since settable expects key first, then value
    // stack: [ "ls"<str>, ls<table> ]
    lua_insert( state, 1 );

    // put ls in the globals table
    // stack: []
    lua_rawset( state, LUA_GLOBALSINDEX );
* /
}
*/
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// HandleLuaFile
//
bool HandleLuaFile(LPCSTR pszFile)
{
    ASSERT_ISSTRING(pszFile);

    char szToken[MAX_LINE_LENGTH] = { 0 };
    
    if (!GetToken(pszFile, szToken, NULL, FALSE) || !szToken[0]) {
        throw std::logic_error("Insufficient tokens. Please specify a script file.");
    }

    try {
        //MessageBox(NULL, pszFile, "Loading script...", MB_TOPMOST | MB_SETFOREGROUND | MB_ICONERROR);
        lua_state->add_script( pszFile );
        //MessageBox(NULL, pszFile, "Script loaded.", MB_TOPMOST | MB_SETFOREGROUND | MB_ICONERROR);

    }
    catch (std::exception& e) {
        Error("Error in Script.\nFile: %s\n\nDescription: %s",
              szToken, e.what());
    }

    return true;
}

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// HandleLuaPackageDir
//
bool HandleLuaPackageDir( LPCSTR pszDir )
{
    ASSERT_ISSTRING( pszDir );

    char szToken[ MAX_LINE_LENGTH ] = { 0 };

    if ( !GetToken( pszDir, szToken, NULL, FALSE ) || !szToken[ 0 ] ) {
        throw std::logic_error( "Insufficient tokens. Please specify a lib directory." );
    }

    try {
        lua_State* state = lua_state->get_lua_state();
        lua_pushstring( state, "LUA_PATH" );
        lua_rawget( state, LUA_GLOBALSINDEX );
        std::string old_path = lua_tostring( state, -1 );
        std::string path = pszDir;
        path = path.substr( 1, path.length() - 2 );
        path += "?.lua;";
        path += pszDir;
        path += "?.lua.dll;";
        path += old_path;
        lua_pushstring( state, "LUA_PATH" );
        lua_pushstring( state, path.c_str() );
        lua_rawset( state, LUA_GLOBALSINDEX );
    }
    catch ( std::exception& e ) {
        Error( "Error adding lib dir.\nDir: %s\n\nDescription: %s",
                szToken, e.what() );
    }

    return true;
}

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// LuaBangHandler
//
void LuaBangHandler( HWND hSender, LPCSTR szCommand, LPCSTR szArgs )
{
    // get lua function name
    std::string lua_func = szCommand;
    lua_func = "bang_" + lua_func.substr( 1, std::string::npos );

    // parse args
    //std::vector< std::string > arg_list;
    // to be implemented later: 
    //      parse out quote delimited strings first
    //      hmmm... trigger( char* name, vector args ) should parse out bools

    lua_state->trigger( lua_func.c_str(), szArgs );
}

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// LuaExecBangHandler
//
void LuaExecBangHandler( HWND hSender, LPCSTR szCommand, LPCSTR szArgs )
{
    lua_State* state = lua_state->get_lua_state();
    std::string buf( szArgs );
    luaL_loadbuffer( state, buf.c_str(), buf.length(), szCommand );
    if ( lua_pcall( state, 0, 0, 0 ) ) {
        Error( lua_tostring( state, -1 ) );
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// EnumSettings
//
bool EnumSettings(LPCSTR pszFile, const std::string& sConfig, ConfigCallback fnCallback)
{
    FILE* f = ::LCOpen(pszFile);

    if (!f) {
        char szWhat[MAX_LINE_LENGTH] = { 0 };

        StringCchPrintf(szWhat, MAX_LINE_LENGTH, "Could not open \"%s\"",
            pszFile ? pszFile : "step.rc");

        throw std::runtime_error(szWhat);
    }

    bool bReturn = true;

    try {
        char szLine[MAX_LINE_LENGTH] = { 0 };

        while (LCReadNextConfig(f, sConfig.c_str(), szLine, MAX_LINE_LENGTH)) {
            LPCSTR pszValue = NULL;
            GetToken(szLine, NULL, &pszValue, FALSE);

            if (pszValue) {
                if (!fnCallback(pszValue)) {
                    bReturn = false;
                    break;
                }
            }
        }
    }
    catch (...) {
        LCClose(f);
        throw;
    }

    LCClose(f);

    return bReturn;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ReadConfig
//
void ReadConfig()
{
//    EnumSettings(NULL, "*Lua", HandleLua);
    EnumSettings( NULL, "*LuaPackageDir", HandleLuaPackageDir );
    EnumSettings( NULL, "*LuaFile", HandleLuaFile );
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ClearConfig
//
void ClearConfig()
{
}

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// SetGlobals
//
void SetGlobals()
{
    // get the directory that the module is stored in. nlm should put the module
    // and all of its lua scripts (ending with .lua.dll) in the same folder,
    // so we can use this in lua to 'require()' desired libraries.
    char dll_path[ MAX_LINE_LENGTH ] = { 0 };
    GetModuleFileName( GetModuleHandle( "lslua" ), dll_path, MAX_LINE_LENGTH );
    std::string path( dll_path );
    // .dll extension required for proper working of nlm
    path = path.substr( 0, path.rfind( '\\' ) ) + "\\?.lua.dll;" +
           path.substr( 0, path.rfind( '\\' ) ) + "\\?.lua";

    lua_State* state = lua_state->get_lua_state();

    // set path lua looks in for require()
    lua_pushstring( state, "LUA_PATH" );
    lua_pushstring( state, path.c_str() );
    lua_rawset( state, LUA_GLOBALSINDEX );

    //make the lslua version accessible through LSLUA_VERSION
    lua_pushstring( state, "LSLUA_VERSION" );
    lua_pushstring( state, rcsRevision );
    lua_rawset( state, LUA_GLOBALSINDEX );
}

void RegisterBangs()
{
    // register bangs by iterating through the list of functions in the
    // lua global symbol table for functions that start with "bang_"
    lua_State* state = lua_state->get_lua_state();
    lua_pushnil( state );
    while ( lua_next( state, LUA_GLOBALSINDEX ) ) {
        if ( lua_isfunction( state, -1 ) ) {
            if ( lua_isstring( state, -2 ) ) {
                std::string funcname = lua_tostring( state, -2 );
                if ( funcname.substr( 0, 5 ) == "bang_" ) {
                    funcname = "!" + funcname.substr( 5, std::string::npos );
                    AddBangCommandEx( funcname.c_str(), LuaBangHandler );
                    //MessageBox( NULL, funcname.c_str(), "Registering bang...", NULL );
                    loaded_bangs.push_back( funcname );
                }
            }
        }
        lua_pop( state, 1 );
    }
    AddBangCommandEx( "!LuaExec", LuaExecBangHandler );
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// initModuleEx
//
int initModuleEx(HWND hLiteStep, HINSTANCE hInstance, LPCSTR szLSpath)
{
    int nReturn = 0;

    try {
        lua_state = new Lua;
        SetGlobals();
        ReadConfig();
        RegisterBangs();
    }
    catch (std::bad_alloc&) {
        Error("Bad allocation during initialization.\nOut of memory?");

        nReturn = 2;
    }
    catch (std::exception& e) {
        Error(e.what());

        nReturn = 1;
    }

    return nReturn;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// quitModule
//
void quitModule(HINSTANCE hInstance)
{
    //MessageBox( NULL, "quitModule", "", NULL );
    ClearConfig();

    for ( std::vector< std::string >::const_iterator vit = loaded_bangs.begin(); vit != loaded_bangs.end(); vit++ ) {
        //MessageBox( NULL, vit->c_str(), "Removing bang...", NULL );
        RemoveBangCommand( vit->c_str() );
    }
    if ( lua_state ) {
        //MessageBox( NULL, "deleting lua_state", "", NULL );
        delete lua_state;
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// DllMain
//
// DLL entry point
//
BOOL APIENTRY DllMain(HINSTANCE hInstance, DWORD dwReason,
                      LPVOID /* pvReserved */)
{
    switch (dwReason)
    {
        case DLL_PROCESS_ATTACH:
        {
            DisableThreadLibraryCalls(hInstance);
        }
        break;
        
        case DLL_PROCESS_DETACH:
        {
        }
        break;
    
        default:
        break;
    }
    
    return TRUE;
}
